package jobqueue

import (
	"encoding/json"
	"fmt"
	"gopkg.in/yaml.v3"
	"io"
	"os"
	"strings"
	"time"
)

type JobImport struct {
	ID           string          `yaml:"id" json:"id"`
	Priority     int             `yaml:"priority" json:"priority"`
	Timeout      time.Duration   `yaml:"timeout" json:"timeout"`
	MaxRetries   uint            `yaml:"maxRetries" json:"maxRetries"`
	RetryDelay   time.Duration   `yaml:"retryDelay" json:"retryDelay"`
	Dependencies []string        `yaml:"dependencies" json:"dependencies,omitempty"`
	Runnable     RunnableImport  `yaml:"runnable" json:"runnable"`
	Scheduler    SchedulerImport `yaml:"scheduler" json:"scheduler,omitempty"`
}

type RunnableImport struct {
	Type string         `yaml:"type" json:"type"`
	Data map[string]any `yaml:"data,omitempty" json:"data,omitempty"`
}

type SchedulerImport struct {
	Type     string        `yaml:"type" json:"type"`
	Interval time.Duration `yaml:"interval,omitempty" json:"interval,omitempty"`
	Spec     string        `yaml:"spec,omitempty" json:"spec,omitempty"`
	Delay    time.Duration `yaml:"delay,omitempty" json:"delay,omitempty"`
	Event    string        `yaml:"event,omitempty" json:"event,omitempty"`
}

func ReadYAML(r io.Reader) ([]JobImport, error) {
	var jobs []JobImport
	decoder := yaml.NewDecoder(r)
	if err := decoder.Decode(&jobs); err != nil {
		return nil, err
	}
	return jobs, nil
}

func ReadJSON(r io.Reader) ([]JobImport, error) {
	var jobs []JobImport
	decoder := json.NewDecoder(r)
	if err := decoder.Decode(&jobs); err != nil {
		return nil, err
	}
	return jobs, nil
}

func ReadYAMLFile(filePath string) ([]JobImport, error) {
	file, err := os.Open(filePath)
	if err != nil {
		return nil, err
	}
	defer file.Close()
	return ReadYAML(file)
}

func ReadJsonFile(filePath string) ([]JobImport, error) {
	file, err := os.Open(filePath)
	if err != nil {
		return nil, err
	}
	defer file.Close()
	return ReadJSON(file)
}

func CreateGenericJobFromImport[T any](jobImport JobImport, runner Runnable[T]) GenericJob {
	return &Job[T]{
		id:         JobID(jobImport.ID),
		priority:   Priority(jobImport.Priority),
		timeout:    jobImport.Timeout,
		maxRetries: jobImport.MaxRetries,
		RetryDelay: jobImport.RetryDelay,
		runner:     runner,
	}
}

func CreateJobAndSchedulerFromImport(jobImport JobImport, manager *Manager) (GenericJob, Scheduler, error) {

	var job GenericJob

	rType := strings.ToLower(jobImport.Runnable.Type)

	runnableData := make(map[string]interface{})
	for k, v := range jobImport.Runnable.Data {
		runnableData[strings.ToLower(k)] = v
	}

	switch rType {
	case "dummy":

		runner, err := NewDummyRunnableFromMap(runnableData)
		if err != nil {
			return nil, nil, err
		}

		job = CreateGenericJobFromImport[DummyResult](jobImport, runner)

	case "counter":

		runner, err := NewCounterRunnableFromMap(runnableData)
		if err != nil {
			return nil, nil, err
		}

		job = CreateGenericJobFromImport[CounterResult](jobImport, runner)

	case "fileoperation":
		runner, err := NewFileOperationRunnableFromMap(runnableData)
		if err != nil {
			return nil, nil, err
		}

		job = CreateGenericJobFromImport[FileOperationResult](jobImport, runner)

	case "db":
		runner, err := NewDBRunnableFromMap(runnableData)
		if err != nil {
			return nil, nil, err
		}

		job = CreateGenericJobFromImport[DBResult](jobImport, runner)

	case "http":
		runner, err := NewHTTPRunnableFromMap(runnableData)
		if err != nil {
			return nil, nil, err
		}

		job = CreateGenericJobFromImport[HTTPResult](jobImport, runner)

	case "mail":
		runner, err := NewMailRunnableFromMap(runnableData)
		if err != nil {
			return nil, nil, err
		}

		job = CreateGenericJobFromImport[MailResult](jobImport, runner)

	case "sftp":
		runner, err := NewSFTPRunnableFromMap(runnableData)
		if err != nil {
			return nil, nil, err
		}

		job = CreateGenericJobFromImport[SFTPResult](jobImport, runner)

	case "shell":
		runner, err := NewShellRunnableFromMap(runnableData)
		if err != nil {
			return nil, nil, err
		}

		job = CreateGenericJobFromImport[ShellResult](jobImport, runner)

	default:
		return nil, nil,
			fmt.Errorf("%w: %s, available types: dummy, counter, fileoperation, db, http, mail, sftp, shell",
				ErrUnknownRunnableType, rType)

	}

	sType := strings.ToLower(jobImport.Scheduler.Type)

	scheduleData := make(map[string]interface{})
	for k, v := range jobImport.Runnable.Data {
		scheduleData[strings.ToLower(k)] = v
	}

	var scheduler Scheduler
	switch sType {
	case "interval":
		scheduler = &IntervalScheduler{Interval: jobImport.Scheduler.Interval}

	case "cron":
		scheduler = &CronScheduler{
			Spec: jobImport.Scheduler.Spec,
		}

		if manager != nil {
			scheduler.(*CronScheduler).cron = manager.GetCronInstance()
		}

	case "delay":
		scheduler = &DelayScheduler{Delay: jobImport.Scheduler.Delay}

	case "event":
		scheduler = &EventScheduler{Event: EventName(jobImport.Scheduler.Event)}

	default:
		return nil, nil, ErrUnknownSchedulerType
	}

	return job, scheduler, nil
}

// LoadJobsAndScheduleFromFile read jobs from a file and schedule them. (json/yaml)
func LoadJobsAndScheduleFromFile(filePath string, manager *Manager) error {

	var err error
	var imp []JobImport

	if filePath[len(filePath)-4:] == "json" {
		imp, err = ReadJsonFile(filePath)
	} else if filePath[len(filePath)-4:] == "yaml" {
		imp, err = ReadYAMLFile(filePath)
	} else {
		return ErrUnknownFormat
	}

	if err != nil {
		return err
	}

	for _, imp := range imp {
		job, scheduler, err := CreateJobAndSchedulerFromImport(imp, manager)
		if err != nil {
			return err
		}

		err = manager.ScheduleJob(job, scheduler)
		if err != nil {
			return err
		}
	}

	return nil

}

// ImportJobsAndSchedule lädt Jobs aus einem Reader und plant sie ein.
func ImportJobsAndSchedule(reader io.Reader, format string, manager *Manager) error {
	var err error
	var imp []JobImport

	// format to lowercase
	format = strings.ToLower(format)

	if format == "json" {
		imp, err = ReadJSON(reader)
	} else if format == "yaml" {
		imp, err = ReadYAML(reader)
	} else {
		return fmt.Errorf("%w: %s", ErrUnknownFormat, format)
	}

	if err != nil {
		return err
	}

	for _, imp := range imp {
		job, scheduler, err := CreateJobAndSchedulerFromImport(imp, manager)
		if err != nil {
			return err
		}

		err = manager.ScheduleJob(job, scheduler)
		if err != nil {
			return err
		}
	}

	return nil
}
