package jobqueue

import (
	"fmt"
	"os"
)

func NewFileOperationRunnableFromMap(data map[string]interface{}) (*FileOperationRunnable, error) {
	operation, ok := data["operation"].(string)
	if !ok {
		return nil, fmt.Errorf("%w: Invalid Operation: %v", ErrInvalidData, data["operation"])
	}

	filePath, ok := data["filepath"].(string)
	if !ok {
		return nil, fmt.Errorf("%w: Invalid FilePath: %v", ErrInvalidData, data["filepath"])
	}

	content, ok := data["content"].(string) // Optional, so no error check
	if !ok {
		return nil, fmt.Errorf("%w: Invalid Content: %v", ErrInvalidData, data["content"])
	}

	return &FileOperationRunnable{
		Operation: operation,
		FilePath:  filePath,
		Content:   content,
	}, nil
}

type FileOperationResult struct {
	Success bool
	Content string // Optional, je nach Operation
}

const (
	FileOperationRead   = "read"
	FileOperationWrite  = "write"
	FileOperationDelete = "delete"
	FileOperationAppend = "append"
	FileOperationCreate = "create"
)

type FileOperationRunnable struct {
	Operation string // z.B. "read", "write", "delete"
	FilePath  string
	Content   string // Optional, je nach Operation
}

func (f *FileOperationRunnable) Run() (RunResult[FileOperationResult], error) {
	switch f.Operation {
	case FileOperationRead:
		content, err := os.ReadFile(f.FilePath)
		if err != nil {
			return RunResult[FileOperationResult]{Status: ResultStatusFailed}, err
		}
		return RunResult[FileOperationResult]{
			Status: ResultStatusSuccess,
			Data: FileOperationResult{
				Success: true,
				Content: string(content),
			},
		}, nil
	case FileOperationWrite:
		err := os.WriteFile(f.FilePath, []byte(f.Content), 0644)
		if err != nil {
			return RunResult[FileOperationResult]{Status: ResultStatusFailed}, err
		}
		return RunResult[FileOperationResult]{
			Status: ResultStatusSuccess,
			Data: FileOperationResult{
				Success: true,
			},
		}, nil
	case FileOperationDelete:
		err := os.Remove(f.FilePath)
		if err != nil {
			return RunResult[FileOperationResult]{Status: ResultStatusFailed}, err
		}
		return RunResult[FileOperationResult]{
			Status: ResultStatusSuccess,
			Data: FileOperationResult{
				Success: true,
			},
		}, nil

	case FileOperationAppend:
		fp, err := os.OpenFile(f.FilePath, os.O_APPEND|os.O_CREATE|os.O_WRONLY, 0644)
		if err != nil {
			return RunResult[FileOperationResult]{Status: ResultStatusFailed}, err
		}

		defer fp.Close()

		if _, err := fp.WriteString(f.Content); err != nil {
			return RunResult[FileOperationResult]{Status: ResultStatusFailed}, err
		}

		return RunResult[FileOperationResult]{
			Status: ResultStatusSuccess,
			Data: FileOperationResult{
				Success: true,
			},
		}, nil

	case FileOperationCreate:
		f, err := os.Create(f.FilePath)
		if err != nil {
			return RunResult[FileOperationResult]{Status: ResultStatusFailed}, err
		}
		defer f.Close()

		return RunResult[FileOperationResult]{
			Status: ResultStatusSuccess,
			Data: FileOperationResult{
				Success: true,
			},
		}, nil

	default:
		return RunResult[FileOperationResult]{Status: ResultStatusFailed}, ErrUnsupportedFileOption
	}
}
